import os
from typing import Optional, Dict, Any

import click
from ruamel import yaml
from ruamel.yaml import Loader

from asphalt.core.runner import run_application, policies
from asphalt.core.utils import merge_config, qualified_name


@click.group()
def main():
    pass  # pragma: no cover


@main.command(help='Read one or more configuration files and start the application.')
@click.argument('configfile', type=click.File(), nargs=-1, required=True)
@click.option('--unsafe', is_flag=True, default=False,
              help='use unsafe mode when loading YAML (enables markup extensions)')
@click.option('-l', '--loop', type=click.Choice(policies.names),
              help='alternate event loop policy')
@click.option('-s', '--service', type=str,
              help='service to run (if the configuration file contains multiple services)')
def run(configfile, unsafe: bool, loop: Optional[str], service: Optional[str]):
    # Read the configuration from the supplied YAML files
    config = {}  # type: Dict[str, Any]
    for path in configfile:
        config_data = yaml.load(path, Loader=Loader) if unsafe else yaml.safe_load(path)
        assert isinstance(config_data, dict), 'the document root element must be a dictionary'
        config = merge_config(config, config_data)

    # Override the event loop policy if specified
    if loop:
        config['event_loop_policy'] = loop

    services = config.pop('services', {})
    if not isinstance(services, dict):
        raise click.ClickException('The "services" key must be a dict, not {}'.format(
            qualified_name(services)))

    # If "component" was defined, use that as the default service if one has not been defined yet
    if 'component' in config:
        component = config.pop('component')
        services.setdefault('default', dict(component=component))

    # Try to figure out which service to launch
    service = service or os.getenv('ASPHALT_SERVICE')
    if len(services) == 0:
        raise click.ClickException('No services have been defined')
    elif service:
        try:
            service_config = services[service]
        except KeyError:
            raise click.ClickException(
                'Service {!r} has not been defined'.format(service)) from None
    elif len(services) == 1:
        service_config = next(iter(services.values()))
    elif 'default' in services:
        service_config = services['default']
    else:
        raise click.ClickException(
            'Multiple services present in configuration file but no default service has been '
            'defined and no service was explicitly selected with -s / --service')

    # Merge the service-level configuration with the top level one
    config = merge_config(config, service_config)

    # Start the application
    run_application(**config)



from pathlib import Path
from unittest.mock import patch

import pytest
from click.exceptions import ClickException
from click.testing import CliRunner

from asphalt.core import cli, Component, Context


class DummyComponent(Component):
    def __init__(self, dummyval1=None, dummyval2=None):
        self.dummyval1 = dummyval1
        self.dummyval2 = dummyval2

    def start(self, ctx: Context):
        pass


@pytest.fixture
def runner():
    return CliRunner()


@pytest.mark.parametrize('loop', [None, 'uvloop'], ids=['default', 'override'])
@pytest.mark.parametrize('unsafe', [False, True], ids=['safe', 'unsafe'])
def test_run(runner, unsafe, loop):
    if unsafe:
        component_class = '!!python/name:{0.__module__}.{0.__name__}'.format(DummyComponent)
    else:
        component_class = '{0.__module__}:{0.__name__}'.format(DummyComponent)

    config = """\
---
event_loop_policy: bogus
component:
  type: {}
  dummyval1: testval
logging:
  version: 1
  disable_existing_loggers: false
""".format(component_class)
    args = ['test.yml']
    if unsafe:
        args.append('--unsafe')
    if loop:
        args.extend(['--loop', loop])

    with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
        Path('test.yml').write_text(config)
        result = runner.invoke(cli.run, args)

        assert result.exit_code == 0
        assert run_app.call_count == 1
        args, kwargs = run_app.call_args
        assert len(args) == 0
        assert kwargs == {
            'component': {
                'type': DummyComponent if unsafe else component_class,
                'dummyval1': 'testval'
            },
            'event_loop_policy': loop or 'bogus',
            'logging': {'version': 1, 'disable_existing_loggers': False}
        }


def test_run_multiple_configs(runner):
    component_class = '{0.__module__}:{0.__name__}'.format(DummyComponent)
    config1 = """\
---
component:
  type: {}
  dummyval1: testval
logging:
  version: 1
  disable_existing_loggers: false
""".format(component_class)
    config2 = """\
---
component:
  dummyval1: alternate
  dummyval2: 10
"""

    with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
        Path('conf1.yml').write_text(config1)
        Path('conf2.yml').write_text(config2)
        result = runner.invoke(cli.run, ['conf1.yml', 'conf2.yml'])

        assert result.exit_code == 0
        assert run_app.call_count == 1
        args, kwargs = run_app.call_args
        assert len(args) == 0
        assert kwargs == {
            'component': {'type': component_class, 'dummyval1': 'alternate', 'dummyval2': 10},
            'logging': {'version': 1, 'disable_existing_loggers': False}
        }


class TestServices:
    def write_config(self):
        Path('config.yml').write_text("""\
---
max_threads: 15
services:
  server:
    max_threads: 30
    component:
      type: myproject.server.ServerComponent
      components:
        wamp: &wamp
          host: wamp.example.org
          port: 8000
          tls: true
          auth_id: serveruser
          auth_secret: serverpass
        mailer:
          backend: smtp
  client:
    component:
      type: myproject.client.ClientComponent
      components:
        wamp:
          <<: *wamp
          auth_id: clientuser
          auth_secret: clientpass
logging:
  version: 1
  disable_existing_loggers: false
""")

    @pytest.mark.parametrize('service', ['server', 'client'])
    def test_run_service(self, runner, service):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            self.write_config()
            result = runner.invoke(cli.run, ['-s', service, 'config.yml'])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            if service == 'server':
                assert kwargs == {
                    'max_threads': 30,
                    'component': {
                        'type': 'myproject.server.ServerComponent',
                        'components': {
                            'wamp': {
                                'host': 'wamp.example.org',
                                'port': 8000,
                                'tls': True,
                                'auth_id': 'serveruser',
                                'auth_secret': 'serverpass'
                            },
                            'mailer': {'backend': 'smtp'}
                        }
                    },
                    'logging': {'version': 1, 'disable_existing_loggers': False}
                }
            else:
                assert kwargs == {
                    'max_threads': 15,
                    'component': {
                        'type': 'myproject.client.ClientComponent',
                        'components': {
                            'wamp': {
                                'host': 'wamp.example.org',
                                'port': 8000,
                                'tls': True,
                                'auth_id': 'clientuser',
                                'auth_secret': 'clientpass'
                            }
                        }
                    },
                    'logging': {'version': 1, 'disable_existing_loggers': False}
                }

    def test_service_not_found(self, runner):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            self.write_config()
            result = runner.invoke(cli.run, ['-s', 'foobar', 'config.yml'])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == "Error: Service 'foobar' has not been defined\n"

    def test_no_service_selected(self, runner):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            self.write_config()
            result = runner.invoke(cli.run, ['config.yml'])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == (
                'Error: Multiple services present in configuration file but no default service '
                'has been defined and no service was explicitly selected with -s / --service\n')

    def test_bad_services_type(self, runner):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            Path('config.yml').write_text("""\
---
services: blah
logging:
  version: 1
  disable_existing_loggers: false
""")
            result = runner.invoke(cli.run, ['config.yml'])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == 'Error: The "services" key must be a dict, not str\n'

    def test_no_services_defined(self, runner):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            Path('config.yml').write_text("""\
---
services: {}
logging:
  version: 1
  disable_existing_loggers: false
""")
            result = runner.invoke(cli.run, ['config.yml'])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == 'Error: No services have been defined\n'

    def test_run_only_service(self, runner):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            Path('config.yml').write_text("""\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
logging:
  version: 1
  disable_existing_loggers: false
""")
            result = runner.invoke(cli.run, ['config.yml'])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                'component': {'type': 'myproject.client.ClientComponent'},
                'logging': {'version': 1, 'disable_existing_loggers': False}
            }

    def test_run_default_service(self, runner):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            Path('config.yml').write_text("""\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
  default:
    component:
      type: myproject.server.ServerComponent
logging:
  version: 1
  disable_existing_loggers: false
""")
            result = runner.invoke(cli.run, ['config.yml'])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                'component': {'type': 'myproject.server.ServerComponent'},
                'logging': {'version': 1, 'disable_existing_loggers': False}
            }

    def test_service_env_variable(self, runner, monkeypatch):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            Path('config.yml').write_text("""\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
  default:
    component:
      type: myproject.server.ServerComponent
logging:
  version: 1
  disable_existing_loggers: false
""")
            monkeypatch.setenv('ASPHALT_SERVICE', 'whatever')
            result = runner.invoke(cli.run, ['config.yml'])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                'component': {'type': 'myproject.client.ClientComponent'},
                'logging': {'version': 1, 'disable_existing_loggers': False}
            }

    def test_service_env_variable_override(self, runner, monkeypatch):
        with runner.isolated_filesystem(), patch('asphalt.core.cli.run_application') as run_app:
            Path('config.yml').write_text("""\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
  default:
    component:
      type: myproject.server.ServerComponent
logging:
  version: 1
  disable_existing_loggers: false
""")
            monkeypatch.setenv('ASPHALT_SERVICE', 'whatever')
            result = runner.invoke(cli.run, ['-s', 'default', 'config.yml'])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                'component': {'type': 'myproject.server.ServerComponent'},
                'logging': {'version': 1, 'disable_existing_loggers': False}
            }
